local tomQuest2 = LibStub("AceAddon-3.0"):NewAddon("tomQuest2", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0", "AceComm-3.0", "AceSerializer-3.0", "LibSink-2.0")
---------------------------------
--- Chargement des Librairies ---
---------------------------------
local L = LibStub("AceLocale-3.0"):GetLocale("tomQuest2")
local LDB = LibStub("LibDataBroker-1.1")
local BC = LibStub("LibBabble-Class-3.0"):GetReverseLookupTable()
local LibTourist = LibStub("LibTourist-3.0")
local LibQTip = LibStub("LibQTip-1.0")
local dataobj = LDB:NewDataObject("tomQuest2", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon"})


------------------------------------
--- Copie des fonctions en local ---
------------------------------------
local CreateFont = CreateFont
local GetDifficultyColor = GetDifficultyColor
local GetNumPartyMembers = GetNumPartyMembers
local GetNumQuestLeaderBoards = GetNumQuestLeaderBoards
local GetNumQuestLogEntries = GetNumQuestLogEntries
local GetQuestLink = GetQuestLink
local GetQuestLogLeaderBoard = GetQuestLogLeaderBoard
local GetQuestLogTitle = GetQuestLogTitle
local GetSubZoneText = GetSubZoneText
local ItemRefTooltip = ItemRefTooltip
local UnitIsDead = UnitIsDead
local UnitName = UnitName
local format = string.format


-------------------------
--- Variables Locales ---
-------------------------
local playerName = UnitName("player")
local _, playerClass = UnitClass("player")
local maxQuests = MAX_QUESTS
local version = 2
local firstCompatibleVersion = 2
local lastIncompatibleVersion = 1
local firstRun = true
local syncParty = false

local questsLogUpdating, questsLogNeedUpdate = false, false
local partyUpdating, partyNeedUpdate = false, false

local party = {}
local quests = {}
local questsObjectives = {}
local questsItems, questsMobs = {}, {}
local questsByZone = {}
local questsZones = {}
local getQuestUidFromId = {}
local getQuestUidFromTitle = {}
local numTrackedQuests = 0

local PARTYQUESTSZONE = L["PARTYQUESTSZONE"]

local questsTracked, questsAutoTracked

local objects_pattern = '^' .. QUEST_OBJECTS_FOUND:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_OBJECTS_FOUND = "%s: %d/%d"
local monsters_pattern = '^' .. QUEST_MONSTERS_KILLED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_MONSTERS_KILLED = "%s slain: %d/%d"
local faction_pattern = '^' .. QUEST_FACTION_NEEDED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_FACTION_NEEDED = "%s: %s / %s"
local player_pattern = '^'.. QUEST_PLAYERS_KILLED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_PLAYERS_KILLED

local DUNGEON = LFG_TYPE_DUNGEON
local HEROIC = DUNGEON_DIFFICULTY2

local reactions = {
  [FACTION_STANDING_LABEL1] = 1,
  [FACTION_STANDING_LABEL1_FEMALE] = 1,
  [FACTION_STANDING_LABEL2] = 2,
  [FACTION_STANDING_LABEL2_FEMALE] = 2,
  [FACTION_STANDING_LABEL3] = 3,
  [FACTION_STANDING_LABEL3_FEMALE] = 3,
  [FACTION_STANDING_LABEL4] = 4,
  [FACTION_STANDING_LABEL4_FEMALE] = 4,
  [FACTION_STANDING_LABEL5] = 5,
  [FACTION_STANDING_LABEL5_FEMALE] = 5,
  [FACTION_STANDING_LABEL6] = 6,
  [FACTION_STANDING_LABEL6_FEMALE] = 6,
  [FACTION_STANDING_LABEL7] = 7,
  [FACTION_STANDING_LABEL7_FEMALE] = 7,
  [FACTION_STANDING_LABEL8] = 8,
  [FACTION_STANDING_LABEL8_FEMALE] = 8,
}

local shorttags = {
  [ELITE] = '+',
  [GROUP] = 'g',
  [PVP] = 'p',
  [RAID] = 'r',
  [DUNGEON] = 'd',
  [HEROIC] = 'd+',
  [DAILY] = "\226\128\162",
}
local tagWeight = {
  [ELITE] = 1,
  [GROUP] = 2,
  [PVP] = 3,
  [RAID] = 4,
  [DUNGEON] = 5,
  [HEROIC] = 6,
  [DAILY] = 7,
}
local questSort = function(a,b)
  local q = quests
  local aa = (q[a].level*4) + (q[a].tag and tagWeight[q[a].tag] or 0)
  local bb = (q[b].level*4) + (q[b].tag and tagWeight[q[b].tag] or 0)
  if aa == bb then
    return q[a].title < q[b].title
  end
  return aa < bb
end

local zoneSort = function(a,b)
  if a == PARTYQUESTSZONE then
    return false
  elseif b == PARTYQUESTSZONE then
    return true
  end
  return a < b
end

local prdb
local zoneDisplay


--------------------------
--- Gestion du Tooltip ---
--------------------------
local tip, tipAnchor
local tracker, trackerAnchor

local tooltipIsShown = false
local plus = [[|TInterface\Buttons\UI-PlusButton-Up:24|t]]
local minus = [[|TInterface\Buttons\UI-MinusButton-Up:24|t]]

local function hideTooltip()
  if MouseIsOver(tip) then return end
  tip:SetScript("OnLeave", nil)
  LibQTip:Release(tip)
  tip = nil
end

local function getColorFromCompletionHEX(percent)
  if percent <= 0 then
    return "ff0000"
  elseif percent >= 1 then
    return "00ff00"
  elseif percent == 0.5 then
    return "ffff00"
  elseif percent < 0.5 then
    percent = percent / 0.5
    return format("%02x%02x%02x",255, percent*255, 0)
  elseif percent < 1 then
    percent = (percent-0.5) / 0.5
    return format("%02x%02x%02x",255 - 255*percent, 255, 0)
  end
end

local function getColorFromCompletionRGB(percent)
  if percent <= 0 then
    return 1, 0, 0
  elseif percent >= 1 then
    return 0, 1, 0
  elseif percent == 0.5 then
    return 1, 1, 0
  elseif percent < 0.5 then
    percent = percent / 0.5
    return 1, percent, 0
  elseif percent < 1 then
    percent = (percent-0.5) / 0.5
    return 1 - percent, 1, 0
  end
end


-----------------------------
--- Quests Tracker Anchor ---
-----------------------------
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)

local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(L["Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)

anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")

anchor:SetScript("OnDragStart", function(self)
  self:StartMoving()
end)

anchor:SetScript("OnDragStop", function(self)
  self:StopMovingOrSizing()
  prdb.tracker.point, prdb.tracker.x, prdb.tracker.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
  trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end)


---------------
--- Options ---
---------------
local defaults = {
  profile = {
    tooltip = {
      headerFontSize = 12,
      normalFontSize = 10,
      maxwidth = 300,
    },
    tracker = {
      point = "CENTER",
      x = 0,
      y = 0,
      enabletracker = true,
      showanchor = true,
      headerFontSize = 12,
      normalFontSize = 10,
      backDropColor = {0.09, 0.09, 0.09, 1},
      borderColor = {1, 1, 1, 1},
      scale = 1,
      maxwidth = 300,
      },
    sinkOptions = {
    },
  },
  char = {
    hidden = {
      ['*'] = {},
    },
  },
}

local options = {
  name = "tomQuest2",
  handler = tomQuest2,
  type = "group",
  args = {
    tooltip = {
      name = "tooltip",
      type = "group",
      args = {
        headerFontSize = {
          name = L["Quests font size"],
          type="range",
          min = 6,
          max = 20,
          step = 1,
          get = function(info)
            return prdb.tooltip.headerFontSize
          end,
          set = function(info, value)
            prdb.tooltip.headerFontSize = value
            if tooltipIsShown then
              tip:SetScript("OnLeave", nil)
              LibQTip:Release(tip)
              tomQuest2:updateTooltip()
            end
          end,
          order=11,
        },
        normalFontSize = { 
          name = L["Quests Objectives font size"],
          type="range",
          min = 6,
          max = 20,
          step = 1,
          get = function(info)
            return prdb.tooltip.normalFontSize
          end,
          set = function(info, value)
            prdb.tooltip.normalFontSize = value
            if tooltipIsShown then
              tip:SetScript("OnLeave", nil)
              LibQTip:Release(tip)
              tomQuest2:updateTooltip()
            end
          end,
          order=21,
        },
        maxWidth = {
          name = L["Quests Tooltip column max width"],
          type="range",
          min = 50,
          max = 600,
          step = 5,
          get = function(info)
            return prdb.tooltip.maxwidth
          end,
          set = function(info, value)
            prdb.tooltip.maxwidth = value
            if prdb.tracker.enabletracker then
              LibQTip:Release(tracker)
              tomQuest2:updateQuestTracker()
            end
          end,
          order=31,
        },
      },
    },
    tracker = {
      name = "tracker",
      type = "group",
      args = {
        enableTracker = {
          name = L["Enable quest tracker"],
          type = "toggle",
          get = function(info)
            return prdb.tracker.enabletracker
          end,
          set = function(info, value)
            prdb.tracker.enabletracker = value
            if prdb.tracker.enabletracker then
              tomQuest2:updateQuestTracker()
              if prdb.tracker.showanchor then
                anchor:Show()
              end
            else
              LibQTip:Release(tracker)
              anchor:Hide()
            end
          end,
          width = "full",
          order = 1,
        },
        showAnchor = {
          name = L["Show quest tracker anchor"],
          type = "toggle",
          get = function(info)
            return prdb.tracker.showanchor
          end,
          set = function(info, value)
            prdb.tracker.showanchor = value
            if prdb.tracker.showanchor and prdb.tracker.enabletracker then
              anchor:Show()
            else
              anchor:Hide()
            end
          end,
          width = "full",
          order = 1,
        },
        headerFontSize = {
          name = L["Quests font size"],
          type="range",
          min = 6,
          max = 20,
          step = 1,
          get = function(info)
            return prdb.tracker.headerFontSize
          end,
          set = function(info, value)
            prdb.tracker.headerFontSize = value
            if prdb.tracker.enabletracker then
              LibQTip:Release(tracker)
              tomQuest2:updateQuestTracker()
            end
          end,
          order=11,
        },
        normalFontSize = { 
          name = L["Quests Objectives font size"],
          type="range",
          min = 6,
          max = 20,
          step = 1,
          get = function(info)
            return prdb.tracker.normalFontSize
          end,
          set = function(info, value)
            prdb.tracker.normalFontSize = value
            if prdb.tracker.enabletracker then
              LibQTip:Release(tracker)
              tomQuest2:updateQuestTracker()
            end
          end,
          order=21,
        },
        backDropColor = {
          name = L["Quests Tracker background color"],
          type = "color",
          hasAlpha = true,
          get = function(info)
            return unpack(prdb.tracker.backDropColor)
          end,
          set = function(info, r, g, b, a)
            prdb.tracker.backDropColor[1] = r
            prdb.tracker.backDropColor[2] = g
            prdb.tracker.backDropColor[3] = b
            prdb.tracker.backDropColor[4] = a
            if prdb.tracker.enabletracker then
              LibQTip:Release(tracker)
              tomQuest2:updateQuestTracker()
            end
          end,
          order = 31,
        },
        borderColor = {
          name = L["Quests Tracker border color"],
          type = "color",
          hasAlpha = true,
          get = function(info)
            return unpack(prdb.tracker.borderColor)
          end,
          set = function(info, r, g, b, a)
            prdb.tracker.borderColor[1] = r
            prdb.tracker.borderColor[2] = g
            prdb.tracker.borderColor[3] = b
            prdb.tracker.borderColor[4] = a
            if prdb.tracker.enabletracker then
              LibQTip:Release(tracker)
              tomQuest2:updateQuestTracker()
            end
          end,
          order = 32,
        },
        scale = {
          name = L["Quests Tracker scale"],
          type="range",
          min = 0,
          max = 1,
          step = 0.1,
          get = function(info)
            return prdb.tracker.scale
          end,
          set = function(info, value)
            prdb.tracker.scale = value
            if prdb.tracker.enabletracker then
              LibQTip:Release(tracker)
              tomQuest2:updateQuestTracker()
            end
          end,
          order=42,
        },
        maxWidth = {
          name = L["Quests Tracker column max width"],
          type="range",
          min = 50,
          max = 600,
          step = 5,
          get = function(info)
            return prdb.tracker.maxwidth
          end,
          set = function(info, value)
            prdb.tracker.maxwidth = value
            if prdb.tracker.enabletracker then
              LibQTip:Release(tracker)
              tomQuest2:updateQuestTracker()
            end
          end,
          order=42,
        },
      },
    },
  },--args options
} --options


----------------------
--- Initialization ---
----------------------
function tomQuest2:OnInitialize()
  --Initialize Saved Variables
  self.db = LibStub("AceDB-3.0"):New("tomQuest2DB", defaults, "default")
  self.db.profile = self.db.profile or {}
  self.db.char = self.db.char or {}
  self.db.char.questsTracked = self.db.char.questsTracked or {}
  self.db.char.questsAutoTracked = self.db.char.questsAutoTracked or {}
  self.db.char.zoneDisplay = self.db.char.zoneDisplay or {}

  prdb = self.db.profile
  zoneDisplay = self.db.char.zoneDisplay
  questsTracked = self.db.char.questsTracked
  questsAutoTracked = self.db.char.questsAutoTracked

  anchor:SetPoint(prdb.tracker.point, prdb.tracker.x, prdb.tracker.y)
  if not prdb.tracker.showanchor then anchor:Hide() end
  if not prdb.tracker.enabletracker then anchor:Hide() end
  trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"

  LibStub("AceConfig-3.0"):RegisterOptionsTable("tomQuest2", options)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions("tomQuest2","tomQuest2")
  self:SetSinkStorage(prdb.sinkOptions)
  options.args.output = self:GetSinkAce3OptionsDataTable()

  self:RegisterEvent("QUEST_LOG_UPDATE")
  self:RegisterEvent("PARTY_MEMBERS_CHANGED")

  --AutoTrack
  self:RegisterEvent("MINIMAP_ZONE_CHANGED", "zoneChanged")
  self:RegisterEvent("ZONE_CHANGED", "zoneChanged")
  self:RegisterEvent("ZONE_CHANGED_INDOORS", "zoneChanged")
  self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "zoneChanged")

  --Redraw Tooltip if player level up when it's displayed
  self:RegisterEvent("PLAYER_LEVEL_UP")

  --Communications between clients
  self:RegisterComm("tomQv", "commHandlerVersion")
  self:RegisterComm("tomQq", "commHandlerQuestBegin")
  self:RegisterComm("tomQo", "commHandlerObjective")
  self:RegisterComm("tomQr", "commHandlerQuestEnd")
  self:RegisterComm("tomQm", "commHandlerObjectiveRequest")
  self:RegisterComm("tomQn", "commHandlerObjectiveReply")
  self:RegisterComm("tomQs", "commHandlerDescriptionRequest")
  self:RegisterComm("tomQt", "commHandlerDescriptionReply")
  self:RegisterComm("tomQp", "commHandlerObjectiveProgress")
  self:RegisterComm("tomQc", "commHandlerQuestComplete")
  self:RegisterComm("tomQf", "commHandlerQuestFailed")
  self:RegisterComm("tomQl", "commHandlerQuestLost")
  self:RegisterComm("tomQd", "commHandlerSyncComplete")

  -- Hooks
  self:SecureHook("AddQuestWatch")
  self:SecureHook("RemoveQuestWatch")
  self:SecureHook("QuestLogTitleButton_OnClick")
  self:SecureHook("QuestWatch_Update")
  self:HookScript(GameTooltip, "OnTooltipSetUnit")
  self:HookScript(GameTooltip, "OnTooltipSetItem")
  self:RawHook("UIErrorsFrame_OnEvent", true)
end


function tomQuest2:computeQuestStatus(uid)
  quests[uid].pqstatus = ""
  for partyName in pairs(quests[uid].party) do
    if tonumber(quests[uid].party[partyName].complete) == 1 or tonumber(quests[uid].party[partyName].complete) == -1 then
      if partyName == playerName then
        quests[uid].pqstatus = format("%s%s", quests[uid].pqstatus, tonumber(quests[uid].party[partyName].complete) == 1 and L["Finished"] or L["Failed"])
      else
        quests[uid].pqstatus = format("|cff%s%s|r: %s%s %s", party[partyName].color or "ffffff", partyName:sub(1,3), tonumber(quests[uid].party[partyName].complete) == 1 and L["Finished"] or L["Failed"], quests[uid].party[playerName] and "," or "", quests[uid].pqstatus)
      end
    end
  end
end

function tomQuest2:computeQuestObjectiveStatus(uid, o)
  questsObjectives[uid][o].qostatus = ""
  local totalgot = 0
  for partyName in pairs(questsObjectives[uid][o].party) do
    totalgot = totalgot + questsObjectives[uid][o].party[partyName].ngot
    if partyName == playerName then
      questsObjectives[uid][o].qostatus = format("%s|cff%s%s|r", questsObjectives[uid][o].qostatus, getColorFromCompletionHEX(questsObjectives[uid][o].party[partyName].ngot/questsObjectives[uid][o].nneed), questsObjectives[uid][o].party[partyName].ngot == questsObjectives[uid][o].nneed and L["d"] or questsObjectives[uid][o].party[partyName].got)
    else
      questsObjectives[uid][o].qostatus = format("|cff%s%s|r: |cff%s%s|r%s %s", party[partyName].color or "ffffff", partyName:sub(1,3), getColorFromCompletionHEX(questsObjectives[uid][o].party[partyName].ngot / questsObjectives[uid][o].nneed), questsObjectives[uid][o].party[partyName].ngot == questsObjectives[uid][o].nneed and L["d"] or questsObjectives[uid][o].party[partyName].got,quests[uid].party[playerName] and "," or "", questsObjectives[uid][o].qostatus)
    end
  end
  questsObjectives[uid][o].qocolor = getColorFromCompletionHEX(totalgot/ quests[uid].playeronquest * questsObjectives[uid][o].nneed)
  questsObjectives[uid][o].qostatus = format("%s/|cff%s%s|r", questsObjectives[uid][o].qostatus, questsObjectives[uid][o].qocolor, questsObjectives[uid][o].need)
end


-----------------------------
--- PARTY_MEMBERS_CHANGED ---
-----------------------------
function tomQuest2:PARTY_MEMBERS_CHANGED()
  if partyUpdating then
    partyNeedUpdate = true
    return
  end
  partyUpdating = true
  local numPartyMembers = GetNumPartyMembers()
  local timeStamp = GetTime()
  local numPartySync = 0
  if numPartyMembers > 0 then
    for i=1, numPartyMembers do
      local partyName = UnitName('party'..i)
      if not party[partyName] then
        party[partyName] = {}
        party[partyName].status = 0
      end
      if not party[partyName].timestamp then
        local _, partyClass = UnitClass(partyName)
        local classColor = RAID_CLASS_COLORS[partyClass]
        party[partyName].color = format("%02x%02x%02x", classColor.r*255, classColor.g*255, classColor.b*255)
        tomQuest2:SendCommMessage("tomQv", tomQuest2:Serialize(version, firstCompatibleVersion), "WHISPER", partyName)
      end
      party[partyName].timestamp = timeStamp
      numPartySync = numPartySync + party[partyName].status
    end
  end

  for p in pairs(party) do
    if party[p].timestamp ~= timeStamp then
      for q in pairs(quests) do
        if quests[q].party[p] then
          quests[q].playeronquest = quests[q].playeronquest - 1
          if questsObjectives[q] then
            for o in pairs(questsObjectives[q]) do
              questsObjectives[q][o].party[p] = nil
              if quests[q].playeronquest == 0 then
                questsObjectives[q][o].party = nil
                questsObjectives[q][o] = nil
              else
                tomQuest2:computeQuestObjectiveStatus(q, o)
              end
            end
          end
          quests[q].party[p] = nil
          if quests[q].playeronquest == 0 then
            questsByZone[quests[q].zone].numquests = questsByZone[quests[q].zone].numquests - 1
            questsByZone[quests[q].zone].quests[q] = nil
            if questsByZone[quests[q].zone].numquests == 0 then
              questsByZone[quests[q].zone].quests = nil
              questsByZone[quests[q].zone] = nil
              questsZones[quests[q].zone] = nil
            end
            if questsObjectives[q] then
              questsObjectives[q] = nil
            end
            quests[q].party = nil
            quests[q] = nil
          else
            tomQuest2:computeQuestStatus(q)
          end
          if questsTracked[q] then
            questsTracked[q] = nil
            numTrackedQuests = numTrackedQuests - 1
            if questsAutoTracked[q] then
              questsAutoTracked[q] = nil
            end
          end
        end
      end
      party[p] = nil
    end
  end

  if numPartySync ~= 0 then
    syncParty = true
  else
    syncParty = false
  end
  partyUpdating = false
  if partyNeedUpdate then
    partyNeedUpdate = false
    tomQuest2:PARTY_MEMBERS_CHANGED()
  end
  
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end


--------------------
--- commHandlers ---
--------------------
function tomQuest2:commHandlerVersion(prefix, message, distribution, sender)
  local success, senderVersion, senderFirstCompatibleVersion = tomQuest2:Deserialize(message)
  if not success then return end
  if senderVersion <= lastIncompatibleVersion then 
    DEFAULT_CHAT_FRAME:AddMessage(format(L["The tomQuest2 release (%d) used by %s is too old to synchronise with your release (%d) (min %d)"], senderVersion, sender, version, firstCompatibleVersion))
    return
  elseif version < senderFirstCompatibleVersion then
    DEFAULT_CHAT_FRAME:AddMessage(format(L["Your tomQuest2 release (%d) is too old to synchorise with release (%d) of %s (min %d)"], version, senderVersion, sender, firstCompatibleVersion))
    return
  elseif version < senderVersion then
    DEFAULT_CHAT_FRAME:AddMessage(format(L["New release available (%d) ask %s where to find it"], senderVersion, sender))
  end
  party[sender] = party[sender] or {}
  party[sender].status = 1
  syncParty = true
  for uid in pairs(quests) do
    if quests[uid].party[playerName] then
      tomQuest2:SendCommMessage("tomQq", tomQuest2:Serialize(uid, quests[uid].title, quests[uid].shorttag, quests[uid].level, quests[uid].numobjectives, quests[uid].party[playerName].complete or 0), "WHISPER", sender, "BULK")
      if questsObjectives[uid] then
        for o in pairs(questsObjectives[uid]) do
          if questsObjectives[uid][o].party[playerName] then
            tomQuest2:SendCommMessage("tomQo", tomQuest2:Serialize(uid, o, questsObjectives[uid][o].desc, questsObjectives[uid][o].qtype, questsObjectives[uid][o].party[playerName].got, questsObjectives[uid][o].need, questsObjectives[uid][o].party[playerName].ngot, questsObjectives[uid][o].nneed), "WHISPER", sender, "BULK")
          end
        end
      end
    end
  end
  tomQuest2:SendCommMessage("tomQd", "d", "WHISPER", sender, "BULK")
end

function tomQuest2:commHandlerQuestBegin(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid, title, shorttag, level, numobjectives, complete = tomQuest2:Deserialize(message)
  if not success then return end
  if uid and not quests[uid] then
    if not questsZones[PARTYQUESTSZONE] then
      questsZones[PARTYQUESTSZONE] = {}
      questsZones[PARTYQUESTSZONE].level = ""
      questsZones[PARTYQUESTSZONE].color = "ffffff"
      questsByZone[PARTYQUESTSZONE] = {}
      questsByZone[PARTYQUESTSZONE].numquests = 0
      questsByZone[PARTYQUESTSZONE].quests = {}
    end
    quests[uid] = {}
    quests[uid].title = title
    quests[uid].level = level or ""
    quests[uid].zone = PARTYQUESTSZONE
    quests[uid].shorttag = shorttag or ""
    quests[uid].taggedname = format("[%s%s] %s", level, shorttag, title)
    quests[uid].numobjectives = numobjectives or 0
    quests[uid].playeronquest = 0
    quests[uid].party = {}
    questsByZone[PARTYQUESTSZONE].numquests = questsByZone[PARTYQUESTSZONE].numquests + 1
    questsByZone[PARTYQUESTSZONE].quests[uid] = true
    tomQuest2:ScheduleTimer("checkQuestObjective", 60, {[1] = uid, [2] = sender})
  end
  quests[uid].playeronquest = quests[uid].playeronquest + 1
  quests[uid].party[sender] = {}
  quests[uid].party[sender].complete = complete
  tomQuest2:computeQuestStatus(uid)
end

function tomQuest2:commHandlerObjective(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid, objective, desc, qtype, got, need, ngot, nneed = tomQuest2:Deserialize(message)
  if not success then return end
  questsObjectives[uid] = questsObjectives[uid] or {}
  if not questsObjectives[uid][objective] then
    questsObjectives[uid][objective] = {}
    questsObjectives[uid][objective].desc = desc
    questsObjectives[uid][objective].need = need
    questsObjectives[uid][objective].nneed = nneed
  end
  questsObjectives[uid][objective].party = questsObjectives[uid][objective].party or {}
  questsObjectives[uid][objective].party[sender] = questsObjectives[uid][objective].party[sender] or {}
  questsObjectives[uid][objective].party[sender].got = got
  questsObjectives[uid][objective].party[sender].ngot = ngot
  if qtype == "item" or qtype == "object" then
    questsItems[desc] = questsItems[desc] or {}
    questsItems[desc][uid] = objective
  elseif qtype == "monster" then
    questsMobs[desc] = questsMobs[desc] or {}
    questsMobs[desc][uid] = objective
  end
  tomQuest2:computeQuestObjectiveStatus(uid, objective)
end

function tomQuest2:commHandlerQuestEnd(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid = tomQuest2:Deserialize(message)
  if not success then return end
  if not (quests[uid] and quests[uid].party[sender]) then return end
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:commHandlerObjectiveRequest(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid = tomQuest2:Deserialize(message)
  if not success then return end
  if not (party[sender] and quests[uid].objective)then return end
  tomQuest2:SendCommMessage("tomQn", tomQuest2:Serialize(uid, quests[uid].objective), "WHISPER", sender, "BULK")
end

function tomQuest2:commHandlerObjectiveReply(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid, objective = tomQuest2:Deserialize(message)
  if not success then return end
  if not quests[uid] then return end
  quests[uid].objective = objective
  local subZone = GetSubZoneText()
  if subZone and subZone ~= "" then
    if quests[uid].objective and quests[uid].objective:find(subZone, 1, true) then
      questsAutoTracked[uid] = true
      questsTracked[uid] = true
      numTrackedQuests = numTrackedQuests + 1
    end
  end
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if questsTracked[uid] and prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:commHandlerDescriptionRequest(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid = tomQuest2:Deserialize(message)
  if not success then return end
  if not (party[sender] and quests[uid].description) then return end
  tomQuest2:SendCommMessage("tomQt", tomQuest2:Serialize(uid, quests[uid].description), "WHISPER", sender, "BULK")
end

function tomQuest2:commHandlerDescriptionReply(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid, description = tomQuest2:Deserialize(message)
  if not success then return end
  if not quests[uid] then return end
  quests[uid].description = description
  if not questsTracked[uid] then
    local subZone = GetSubZoneText()
    if subZone and subZone ~= "" then
      if quests[uid].description and quests[uid].description:find(subZone, 1, true) then
        questsAutoTracked[uid] = true
        questsTracked[uid] = true
        numTrackedQuests = numTrackedQuests + 1
      end
    end
  end
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if questsTracked[uid] and prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:commHandlerObjectiveProgress(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid, objective, got, ngot = tomQuest2:Deserialize(message)
  if not success then return end
  if not (quests[uid] and quests[uid].party[sender] and questsObjectives[uid][objective].party[sender]) then return end
  questsObjectives[uid][objective].party[sender].got = got
  questsObjectives[uid][objective].party[sender].ngot = ngot
  if not questsTracked[uid] then
    questsTracked[uid] = true
    numTrackedQuests = numTrackedQuests + 1
  elseif questsAutoTracked[uid] then
    questsAutoTracked[uid] = nil
  end
  tomQuest2:computeQuestObjectiveStatus(uid, objective)
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
  if got == questsObjectives[uid][objective].need and quests[uid].party[sender].complete ~= 1 then
    tomQuest2:Pour(format(L["(%s : %s) %s finished"], sender, quests[uid].title, questsObjectives[uid][objective].desc), getColorFromCompletionRGB(1))
  else
    tomQuest2:Pour(format("(%s : %s) %s : %s / %s", sender, quests[uid].title, questsObjectives[uid][objective].desc, got, questsObjectives[uid][objective].need), getColorFromCompletionRGB(ngot/questsObjectives[uid][objective].nneed))
  end
end

function tomQuest2:commHandlerQuestComplete(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid = tomQuest2:Deserialize(message)
  if not success then return end
  if not (quests[uid] and quests[uid].party[sender]) then return end
  quests[uid].party[sender].complete = 1
  tomQuest2:Pour(format(L["(%s : %s) finished"], sender, quests[uid].title), getColorFromCompletionRGB(1))
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if questsTracked[uid] and prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:commHandlerQuestFailed(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid = tomQuest2:Deserialize(message)
  if not success then return end
  if not (quests[uid] and quests[uid].party[sender]) then return end
  quests[uid].party[sender].complete = -1
  tomQuest2:Pour(format(L["(%s : %s) failed"], sender, quests[uid].title), getColorFromCompletionRGB(0))
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if questsTracked[uid] and prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:commHandlerQuestLost(prefix, message, distribution, sender)
  if not party[sender] then return end
  local success, uid = tomQuest2:Deserialize(message)
  if not success then return end
  if not (quests[uid] and quests[uid].party[sender]) then return end
  quests[uid].playeronquest = quests[uid].playeronquest - 1
  if questsObjectives[uid] then
    for o in pairs(questsObjectives[uid]) do
      questsObjectives[uid][o].party[sender] = nil
      if quests[uid].playeronquest == 0 then
        questsObjectives[uid][o].party = nil
        questsObjectives[uid][o] = nil
      end
    end
  end
  quests[uid].party[sender] = nil
  if quests[uid].playeronquest == 0 then
    questsObjectives[uid] = nil
    quests[uid].party = nil
    questsByZone[quests[uid].zone].numquests = questsByZone[quests[uid].zone].numquests - 1
    questsByZone[quests[uid].zone].quests[uid] = nil
    if questsByZone[quests[uid].zone].numquests == 0 then
      questsByZone[quests[uid].zone].quests = nil
      questsByZone[quests[uid].zone] = nil
      questsZones[quests[uid].zone] = nil
    end
    quests[uid] = nil
    if questsTracked[uid] then
      questsTracked[uid] = nil
      numTrackedQuests = numTrackedQuests - 1
      if questsAutoTracked[uid] then
        questsAutoTracked[uid] = nil
      end
    end
  end
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:commHandlerSyncComplete(prefix, message, distribution, sender)
  if not party[sender] then return end
  if tooltipIsShown then
    tip:SetScript("OnLeave", nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:checkQuestObjective(arg)
  if type(arg) ~= "table" then return end
  if not party[arg[2]] then return end
  if not quests[arg[1]] then return end
  if arg[1] and quests[arg[1]] and not quests[arg[1]].objective then
    tomQuest2:SendCommMessage("tomQm", tomQuest2:Serialize(arg[1]), "WHISPER", arg[2], "BULK")
  end
  if arg[1] and quests[arg[1]] and not quests[arg[1]].description then
    tomQuest2:SendCommMessage("tomQs", tomQuest2:Serialize(arg[1]), "WHISPER", arg[2], "BULK")
  end
end


------------------------
--- QUEST_LOG_UPDATE ---
------------------------
function tomQuest2:QUEST_LOG_UPDATE()
  if questsLogUpdating  then
    questsLogNeedUpdate = true
    return
  end
  questsLogUpdating = true
  local selectionId = GetQuestLogSelection()
  local timeStamp = GetTime()
  local computeQuestStatus, computeQuestObjectiveStatus = false, false
  local numEntries, numQuests = GetNumQuestLogEntries()

  for i in pairs(getQuestUidFromId) do
    getQuestUidFromId[i] = nil
  end
  for t in pairs(getQuestUidFromTitle) do
    getQuestUidFromTitle[t] = nil
  end
  if numEntries > 0 then
    local zone
    for id = 1, numEntries do
      local title, level, tag, group, header, collapsed, complete, daily = GetQuestLogTitle(id)
      complete = complete or 0
      if header then
        zone = title
        if not questsZones[zone] then
          questsZones[zone] = {}
          if string.upper(BC[zone] or "") == playerClass then
            local c = RAID_CLASS_COLORS[playerClass]
            questsZones[zone].color = format("%02x%02x%02x", c.r*255, c.g*255, c.b*255)
            questsZones[zone].minlevel, questsZones[zone].maxlevel, questsZones[zone].level = "", "", ""
          else
            questsZones[zone].minlevel, questsZones[zone].maxlevel = LibTourist:GetLevel(zone)
            if questsZones[zone].minlevel == 0 then
              questsZones[zone].level = ""
              questsZones[zone].color = "ffffff"
            else
              questsZones[zone].level = format("(%s-%s)",questsZones[zone].minlevel, questsZones[zone].maxlevel)
              local r, g, b = LibTourist:GetLevelColor(zone)
              questsZones[zone].color = format("%02x%02x%02x", r*255, g*255, b*255)
            end
          end
          questsByZone[zone] = {}
          questsByZone[zone].numquests = 0
          questsByZone[zone].quests = {}
        end
      else
        local uid = tonumber(string.match(GetQuestLink(id), 'quest:(%d+)'))
        getQuestUidFromId[id] = uid
        getQuestUidFromTitle[title] = uid
        if not quests[uid] then
          quests[uid] = {}
          quests[uid].title = title
          quests[uid].level = level
          quests[uid].tag = tag and tag or daily and DAILY or nil
          if (not group or group == 0) then group = "" end
          quests[uid].group = group
          quests[uid].daily = daily
          quests[uid].playeronquest = 0
          quests[uid].shorttag = format("%s%s%s", tag and shorttags[tag] or "", group, daily and (tag ~= DAILY and shorttags[DAILY] or "") or "")
          quests[uid].taggedname = format("[%s%s] %s", level or "", quests[uid].shorttag or "", title)
          quests[uid].zone = zone
          quests[uid].party = {}
          questsByZone[zone].quests[uid] = true
          questsByZone[zone].numquests = questsByZone[zone].numquests + 1
        elseif quests[uid].zone == PARTYQUESTSZONE then
          quests[uid].zone = zone
          questsByZone[zone].quests[uid] = true
          questsByZone[zone].numquests = questsByZone[zone].numquests + 1
          questsByZone[PARTYQUESTSZONE].quests[uid] = nil
          questsByZone[PARTYQUESTSZONE].numquests = questsByZone[PARTYQUESTSZONE].numquests - 1
          if questsByZone[PARTYQUESTSZONE].numquests == 0 then
            questsByZone[PARTYQUESTSZONE].quests = nil
            questsByZone[PARTYQUESTSZONE] = nil
          end
        end
        quests[uid].id = id
        if not quests[uid].party[playerName] then
          quests[uid].party[playerName] = {}
          quests[uid].party[playerName].complete = complete
          quests[uid].playeronquest = quests[uid].playeronquest + 1
          if not (quests[uid].description or quests[uid].objective) then
            SelectQuestLogEntry(quests[uid].id)
            quests[uid].description, quests[uid].objective = GetQuestLogQuestText()
          end
          quests[uid].numobjectives = GetNumQuestLeaderBoards(id) or 0
          if not questsTracked[uid] then
            local subZone = GetSubZoneText()
            if subZone and subZone ~= "" then
              if (quests[uid].objective and quests[uid].objective:find(subZone, 1, true)) or (quests[uid].description and quests[uid].description:find(subZone, 1, true)) then
                questsAutoTracked[uid] = true
                questsTracked[uid] = true
                numTrackedQuests = numTrackedQuests + 1
              end
            end
          end
          if GetNumPartyMembers() > 0 and syncParty and not firstRun then
            tomQuest2:SendCommMessage("tomQq", tomQuest2:Serialize(uid, title, quests[uid].shorttag, quests[uid].level, quests[uid].numobjectives, quests[uid].party[playerName].complete), "PARTY", "BULK")
          end
        else
          if not firstRun and quests[uid].party[playerName].complete ~= complete then
            if complete == 1 then
              PlaySoundFile([[Interface\Addons\tomQuest2\media\Peasant_work_done.mp3]])
              tomQuest2:Pour(format(L["(%s) : finished"], quests[uid].title), getColorFromCompletionRGB(1))
              if GetNumPartyMembers() > 0 and syncParty then
                tomQuest2:SendCommMessage("tomQc", tomQuest2:Serialize(uid), "PARTY", "BULK")
              end
            elseif complete == -1 then
              PlaySound([[Sound\Interface\igPlayerBind.wav]])
              tomQuest2:Pour(format(L["(%s) : failed"], quests[uid].title), getColorFromCompletionRGB(0))
              if GetNumPartyMembers() > 0 and syncParty then
                tomQuest2:SendCommMessage("tomQf", tomQuest2:Serialize(uid), "PARTY", "BULK")
              end
            end
            computeQuestStatus = true
          end
          quests[uid].party[playerName].complete = complete
        end
        if computeQuestStatus or firstRun then
          computeQuestStatus = false
          tomQuest2:computeQuestStatus(uid)
        end
        if quests[uid].numobjectives > 0 then
          questsObjectives[uid] = questsObjectives[uid] or {}
          for o = 1, quests[uid].numobjectives do
            questsObjectives[uid][o] = questsObjectives[uid][o] or {}
            local need, got, nneed, ngot, desc
            local description, qtype, done = GetQuestLogLeaderBoard(o, id)
            if qtype == "item" or qtype == "object" then
              desc, got, need = description:match(objects_pattern)
              nneed, ngot = need, got
            elseif qtype == "player" then
              desc, got, need = description:match(player_pattern)
              nneed, ngot = need, got
            elseif qtype == "monster" then
              desc, got, need = description:match(monsters_pattern)
              if desc == nil or got == nil or need == nil then
                desc, got, need = description:match(objects_pattern)
              end
              nneed, ngot = need, got
            elseif qtype == "reputation" then
              desc, got, need = description:match(faction_pattern)
              nneed, ngot = reactions[need], reactions[got]
            elseif qtype == "event" then
               desc = description
               got = done and 1 or 0
               need = 1
               nneed, ngot = need, got
            end
            questsObjectives[uid][o].desc = desc
            questsObjectives[uid][o].qtype = qtype
            questsObjectives[uid][o].nneed = nneed
            questsObjectives[uid][o].need = need
            questsObjectives[uid][o].party = questsObjectives[uid][o].party or {}
            if not questsObjectives[uid][o].party[playerName] then
              questsObjectives[uid][o].party[playerName] = {}
              if GetNumPartyMembers() > 0 and syncParty and not firstRun then
                tomQuest2:SendCommMessage("tomQo", tomQuest2:Serialize(uid, o, desc, qtype, got, need, ngot, nneed), "PARTY", "BULK")
              end
              if qtype == "item" or qtype == "object" then
                questsItems[desc] = questsItems[desc] or {}
                questsItems[desc][uid] = o
              elseif qtype == "monster" then
                questsMobs[desc] = questsMobs[desc] or {}
                questsMobs[desc][uid] = o
              end
              computeQuestObjectiveStatus = true
            else
              if questsObjectives[uid][o].party[playerName].ngot ~= ngot then
                if not questsTracked[uid] then
                  questsTracked[uid] = true
                  numTrackedQuests = numTrackedQuests + 1
                elseif questsAutoTracked[uid] then 
                  questsAutoTracked[uid] = nil
                end
                if not firstRun and complete == 0 then
                  if tonumber(ngot) == tonumber(nneed) then
                    PlaySound[[Sound\Creature\Peon\PeonReady1.wav]]
                    tomQuest2:Pour(format(L["(%s) %s : finished"], title, desc), getColorFromCompletionRGB(1))
                  else
                    tomQuest2:Pour(format("(%s) %s : %s/%s", title, desc , got, need), getColorFromCompletionRGB(ngot/nneed))
                  end
                end
                if GetNumPartyMembers() > 0 and syncParty and not firstRun then
                  tomQuest2:SendCommMessage("tomQp", tomQuest2:Serialize(uid, o, got, ngot), "PARTY", "BULK")
                end
                computeQuestObjectiveStatus = true
              end
            end
            questsObjectives[uid][o].party[playerName].ngot = ngot
            questsObjectives[uid][o].party[playerName].got = got
            if computeQuestObjectiveStatus or firstRun then
              computeQuestObjectiveStatus = false
              tomQuest2:computeQuestObjectiveStatus(uid, o)
            end
          end
        end
        if not quests[uid].timestamp and GetNumPartyMembers() > 0 and syncParty and not firstRun then
          tomQuest2:SendCommMessage("tomQr", tomQuest2:Serialize(uid), "PARTY", "BULK")
        end
        quests[uid].timestamp = timeStamp
      end
    end
  end
  if GetQuestLogSelection() ~= selectionId then
    SelectQuestLogEntry(selectionId)
  end

  if not firstRun then
    for q in pairs(quests) do
      if tonumber(quests[q].timestamp) ~= tonumber(timeStamp) then
        quests[q].playeronquest = tonumber(quests[q].playeronquest) - 1
        if questsObjectives[q] then
          for o in pairs(questsObjectives[q]) do
            questsObjectives[q][o].party[playerName] = nil
            if quests[q].playeronquest == 0 then
              questsObjectives[q][o].party = nil
              questsObjectives[q][o] = nil
            end
          end
        end
        quests[q].party[playerName] = nil
        questsByZone[quests[q].zone].numquests = questsByZone[quests[q].zone].numquests - 1
        questsByZone[quests[q].zone].quests[q] = nil
        if questsByZone[quests[q].zone].numquests == 0 then
          questsByZone[quests[q].zone].quests = nil
          questsByZone[quests[q].zone] = nil
          questsZones[quests[q].zone] = nil
        end
        if quests[q].playeronquest == 0 then
          questsObjectives[q] = nil
          quests[q].party = nil
          quests[q] = nil
        else
          quests[q].zone = PARTYQUESTSZONE
          if not questsByZone[PARTYQUESTSZONE] then
            questsZones[PARTYQUESTSZONE] = {}
            questsZones[PARTYQUESTSZONE].level = ""
            questsZones[PARTYQUESTSZONE].color = "ffffff"
            questsByZone[PARTYQUESTSZONE] = {}
            questsByZone[PARTYQUESTSZONE].numquests = 0
            questsByZone[PARTYQUESTSZONE].quests = {}
          end
          questsByZone[PARTYQUESTSZONE].numquests = questsByZone[PARTYQUESTSZONE].numquests + 1
          questsByZone[PARTYQUESTSZONE].quests[q] = true
        end
        tomQuest2:SendCommMessage("tomQl",tomQuest2:Serialize(q), "PARTY", "ALERT")
        if questsAutoTracked[q] then
          questsAutoTracked[q] = nil
        end
        if questsTracked[q] then
          questsTracked[q] = nil
          numTrackedQuests = numTrackedQuests - 1
        end
      end
    end
  end

  dataobj.text = format("|cff%s%d / %d|r", getColorFromCompletionHEX(numQuests/maxQuests), numQuests, maxQuests)
  questsLogUpdating = false
  if questsLogNeedUpdate then
    questsLogNeedUpdate = false
    tomQuest2:UPDATE_QUEST_LOG()
  end
  if firstRun then
    for q in pairs(questsTracked) do
      if not quests[q] then
        questsTracked[q] = nil
      else
        numTrackedQuests = numTrackedQuests + 1
      end
    end
    tomQuest2:PARTY_MEMBERS_CHANGED()
  end
  if tooltipIsShown then
    tip:SetScript("OnLeave",nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
  firstRun = false
end


------------------
--- WOW Events ---
------------------
function tomQuest2:zoneChanged()
  if UnitIsDead("player") then return end
  local subZone = GetSubZoneText()
  for uid in pairs(questsAutoTracked) do
    questsAutoTracked[uid] = nil
    questsTracked[uid] = nil
    numTrackedQuests = numTrackedQuests - 1
  end
  if subZone and subZone ~= "" then
    for uid in pairs(quests) do
      if not questsTracked[uid] and (quests[uid].objective and quests[uid].objective:find(subZone, 1, true)) or (quests[uid].description and quests[uid].description:find(subZone, 1, true)) then
        questsAutoTracked[uid] = true
        questsTracked[uid] = true
        numTrackedQuests = numTrackedQuests + 1
      end
    end
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:PLAYER_LEVEL_UP()
  for z in pairs(questsZones) do
    if questsZones[z].minlevel ~= 0 then
      local r, g, b = LibTourist:GetLevelColor(z)
      questsZones[z].color = string.format("%02x%02x%02x", r*255, g*255, b*255)
    end
  end
  if tooltipIsShown then
    tip:SetScript("OnLeave",nil)
    LibQTip:Release(tip)
    tomQuest2:updateTooltip()
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end


--------------------
--- Secure Hooks ---
--------------------
function tomQuest2:AddQuestWatch(id)
  local uid = getQuestUidFromId[id]
  if uid and not questsTracked[uid] then
    questsTracked[uid] = true
    numTrackedQuests = numTrackedQuests + 1
  elseif uid and questsAutoTracked[uid] then
    questsAutoTracked[uid] = nil
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:RemoveQuestWatch(id)
  local uid = getQuestUidFromId[id]
  if uid and questsTracked[uid] then
    questsTracked[uid] = nil
    numTrackedQuests = numTrackedQuests - 1
    if questsAutoTracked[uid] then
      questsAutoTracked[uid] = nil
    end
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:QuestLogTitleButton_OnClick(button)
  if IsModifiedClick("QUESTWATCHTOGGLE") then
    local uid = getQuestUidFromId[this:GetID() + FauxScrollFrame_GetOffset(QuestLogListScrollFrame)]
    if not uid then return end
    if questsTracked[uid] then
      questsTracked[uid] = nil
      numTrackedQuests = numTrackedQuests - 1
      if questsAutoTracked[uid] then
        questsAutoTracked[uid] = nil
      end
    else
      questsTracked[uid] = true
      numTrackedQuests = numTrackedQuests + 1
    end
    if prdb.tracker.enabletracker then
      LibQTip:Release(tracker)
      tomQuest2:updateQuestTracker()
    end
  end
end

function tomQuest2:QuestWatch_Update()
  if prdb.tracker.enabletracker then
    QuestWatchFrame:Hide()
  end
end

function tomQuest2:ToggleQuestWatch(uid)
  if questsTracked[uid] then
    questsTracked[uid] = nil
    numTrackedQuests = numTrackedQuests - 1
    if questsAutoTracked[uid] then
      questsAutoTracked[uid] = nil
    end
  else
    questsTracked[uid] = true
    numTrackedQuests = numTrackedQuests + 1
  end
  if prdb.tracker.enabletracker then
    LibQTip:Release(tracker)
    tomQuest2:updateQuestTracker()
  end
end

function tomQuest2:OnTooltipSetUnit(tooltip, ...)
  local name = tooltip:GetUnit()
  if name and questsMobs[name] then
    for uid in pairs(questsMobs[name]) do
      tooltip:AddDoubleLine(format("|cffffffff%s|r",quests[uid].title), questsObjectives[uid][questsMobs[name][uid]].qostatus)
    end
  end
end

function tomQuest2:OnTooltipSetItem(tooltip, ...)
  local name = tooltip:GetItem()
  if name and questsItems[name] then
    for uid in pairs(questsItems[name]) do
      tooltip:AddDoubleLine(format("|cffffffff%s|r",quests[uid].title), questsObjectives[uid][questsItems[name][uid]].qostatus)
    end
  end
end

local function patternize(s) return '^' .. s:gsub('(%%.)', '(.+)') .. '$' end
local found = patternize(ERR_QUEST_ADD_FOUND_SII) -- "%s: %d/%d"
local item = patternize(ERR_QUEST_ADD_ITEM_SII) -- "%s: %d/%d"
local kill = patternize(ERR_QUEST_ADD_KILL_SII) -- "%s slain: %d/%d"
local complete = patternize(ERR_QUEST_OBJECTIVE_COMPLETE_S) -- "%s (Complete)"
function tomQuest2:UIErrorsFrame_OnEvent(frame, event, msg, ...)
  if event == "UI_INFO_MESSAGE" and (msg:match(found) or msg:match(item) or msg:match(kill) or msg:match(complete)) then 
    return
  end
  return self.hooks.UIErrorsFrame_OnEvent(frame, event, msg, ...)
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
  self.fontString = self:CreateFontString()
  self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, btype, bvalue, maxWidth, indent, anchor)
  self:EnableMouse(true)
  self:SetScript("OnMouseDown", function(frame,button)
    if btype == "zone" then
      zoneDisplay[bvalue] = zoneDisplay[bvalue] == 1 and 0 or 1
      tip:SetScript("OnLeave", nil)
      LibQTip:Release(tip)
      tomQuest2:updateTooltip()
    elseif btype == "quest" then
      if button == "RightButton" then
        if ChatFrameEditBox:IsVisible() then
          if IsShiftKeyDown() and quests[bvalue].party[playerName] and quests[bvalue].id then
            ChatEdit_InsertLink(GetQuestLink(quests[bvalue].id))
          elseif IsControlKeyDown() then
            if quests[bvalue].numobjectives > 0 and quests[bvalue].party[playerName] then
              for o in pairs(questsObjectives[bvalue]) do
                ChatFrameEditBox:Insert(format("{%s %s/%s} ", questsObjectives[bvalue][o].desc, questsObjectives[bvalue][o].party[playerName].got, questsObjectives[bvalue][o].need))
              end
            end
          end
        elseif quests[bvalue].party[playerName] and quests[bvalue].id then
          ItemRefTooltip:ClearLines()
          ShowUIPanel(ItemRefTooltip)
          if not ItemRefTooltip:IsVisible() then
            ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
          end
          local link = GetQuestLink(quests[bvalue].id)
          if link then
            ItemRefTooltip:SetHyperlink(link)
          end
        end
      elseif button == "LeftButton" then
        if IsControlKeyDown() then
          if GetNumPartyMembers() > 0 and quests[bvalue].party[playerName] and quests[bvalue].id then
            local wasSelected = GetQuestLogSelection()
            SelectQuestLogEntry(quests[bvalue].id)
            if GetQuestLogPushable() then
              QuestLogPushQuest()
            end
            SelectQuestLogEntry(wasSelected)
          end
        elseif IsShiftKeyDown() then
          if questsTracked[bvalue] then
            questsTracked[bvalue] = nil
            numTrackedQuests = numTrackedQuests - 1
          else
            questsTracked[bvalue] = true
            numTrackedQuests = numTrackedQuests + 1
          end
          if prdb.tracker.enabletracker then
            LibQTip:Release(tracker)
            tomQuest2:updateQuestTracker()
          end
        elseif quests[bvalue].party[playerName] and quests[bvalue].id then
          if QuestLogFrame:IsVisible() then
            if GetQuestLogSelection() == quests[bvalue].id then
              return HideUIPanel(QuestLogFrame)
            end
          else
            ShowUIPanel(QuestLogFrame)
          end
          local numEntries = GetNumQuestLogEntries()
          if (numEntries > QUESTS_DISPLAYED) then
            if (quests[bvalue].id <= numEntries - QUESTS_DISPLAYED) then
              FauxScrollFrame_SetOffset(QuestLogListScrollFrame, quests[bvalue].id - 1)
              QuestLogListScrollFrameScrollBar:SetValue((quests[bvalue].id - 1) * QUESTLOG_QUEST_HEIGHT)
            else
              FauxScrollFrame_SetOffset(QuestLogListScrollFrame, numEntries - QUESTS_DISPLAYED)
              QuestLogListScrollFrameScrollBar:SetValue((numEntries - QUESTS_DISPLAYED) * QUESTLOG_QUEST_HEIGHT)
            end
          end
          SelectQuestLogEntry(quests[bvalue].id)
          QuestLog_SetSelection(quests[bvalue].id)
          _G["QuestLogTitle"..(quests[bvalue].id - FauxScrollFrame_GetOffset(QuestLogListScrollFrame))]:Click()
        end
      end
    end
  end)
  self:SetScript('OnEnter', function()
    if btype == "quest" and quests[bvalue].party[playerName] then
      local link = GetQuestLink(quests[bvalue].id)
      if link then
        GameTooltip:SetOwner(tooltip, anchor)
        GameTooltip:SetHyperlink(link)
      end
    end
  end)
  self:SetScript('OnLeave', function()
    if btype == "quest" and quests[bvalue].party[playerName] then
      GameTooltip:Hide()
    end
  end)

  indent = indent or 0
  local fs = self.fontString
  fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
  fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
  local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, arg)
  width = width + indent
  if width > maxWidth then
    fs:SetWidth(maxWidth)
    height = fs:GetHeight()
    width = maxWidth
  end
  fs:Show()
  return width, height
end

function cPrototype:ReleaseCell() 
  self:EnableMouse(false) 
  self:SetScript("OnEnter", nil) 
  self:SetScript("OnLeave", nil) 
  self:SetScript("OnMouseDown", nil) 
end


----------------
--- Tooltip! ---
----------------
function tomQuest2:updateTooltip()
  tip = LibQTip:Acquire("tomQuest2ToolTip", 2, "LEFT", "RIGHT")
  tip:Clear()
  tip:EnableMouse(true)
  tip:SetScript("OnLeave",hideTooltip)

  local headerFont = CreateFont("headerFont")
  headerFont:SetFont(GameTooltipText:GetFont(), prdb.tooltip.headerFontSize)
  tip:SetHeaderFont(headerFont)

  local smallFont = CreateFont("smallFont")
  smallFont:SetFont(GameTooltipText:GetFont(), prdb.tooltip.normalFontSize)
  tip:SetFont(smallFont)

  local tempLine
  
  tempLine = tip:AddLine()
  tip:SetCell(tempLine, 1, "tomQuest2", headerFont, "CENTER", 2, cProvider, "other", nil, prdb.tooltip.maxwidth)

  tip:AddLine(" ")
  for zone in pairs(questsZones) do
    zoneDisplay[zone] = zoneDisplay[zone] or 1
    tempLine = tip:AddLine()
    tip:SetCell(tempLine, 1, string.format("%s |cff%s%s %s|r", zoneDisplay[zone] == 0 and plus or minus, questsZones[zone].color, zone, questsZones[zone].level ), headerFont, "LEFT", 2, cProvider, "zone", zone, prdb.tooltip.maxwidth, 0)
    if zoneDisplay[zone] == 1 then
      for uid in pairs(questsByZone[zone].quests) do
        local qc = GetDifficultyColor(quests[uid].level)
        local questColor = format("%02x%02x%02x", qc.r*255, qc.g*255, qc.b*255)
        tempLine = tip:AddLine()
        tip:SetCell(tempLine, 1, format("|cff%s%s %s|r", questColor, quests[uid].taggedname, quests[uid].playeronquest > 1 and format("<%d>",quests[uid].playeronquest) or ""),headerFont, "LEFT", 1, cProvider, "quest", uid, prdb.tooltip.maxwidth, 5, tipAnchor)
        tip:SetCell(tempLine, 2, quests[uid].pqstatus, headerFont, "RIGHT", 1, cProvider, "quest", uid, prdb.tooltip.maxwidth, 0, tipAnchor)
        if questsObjectives[uid] then
          for o in pairs(questsObjectives[uid]) do
            tempLine = tip:AddLine()
            tip:SetCell(tempLine, 1, format("|cff%s%s|r", questsObjectives[uid][o].qocolor, questsObjectives[uid][o].desc), smallFont, "LEFT", 1, cProvider, "quest", uid, prdb.tooltip.maxwidth, 10, tipAnchor)
            tip:SetCell(tempLine, 2, questsObjectives[uid][o].qostatus, smallFont, "RIGHT", 1, cProvider, "quest", uid, prdb.tooltip.maxwidth, 0, tipAnchor)
          end
        else
          tempLine = tip:AddLine()
          tip:SetCell(tempLine, 1, quests[uid].objective, smallFont, "LEFT", 1, cProvider,"quest", uid, prdb.tooltip.maxwidth, 10, tipAnchor)
        end
      end
    end
  end

  tip:AddLine(" ")
  tempLine = tip:AddLine()
  tip:SetCell(tempLine, 1, L["|cffeda55fLeft click|r |cff19ff19Open quests log.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.tooltip.maxwidth)
  tempLine = tip:AddLine()
  tip:SetCell(tempLine, 1, L["|cffeda55fCtrl + Left click|r |cff19ff19Share quest.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.tooltip.maxwidth)
  tempLine = tip:AddLine()
  tip:SetCell(tempLine, 1, L["|cffeda55fShift + Left click|r |cff19ff19Add quest to tracker.|r"] , smallFont, "LEFT", 2, cProvider, "other", nil, prdb.tooltip.maxwidth)
  tempLine = tip:AddLine()
  tip:SetCell(tempLine, 1, L["|cffeda55fRight click|r |cff19ff19Show quest tooltip.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.tooltip.maxwidth)
  tempLine = tip:AddLine()
  tip:SetCell(tempLine, 1, L["|cffeda55fShift + Right click|r |cff19ff19Send quest link to chat.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.tooltip.maxwidth)
  tempLine = tip:AddLine()
  tip:SetCell(tempLine, 1, L["|cffeda55fCtrl + Right click|r |cff19ff19Send quest progress to chat.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.tooltip.maxwidth)
  tip:Show()
end

function dataobj.OnEnter(self)
  tomQuest2:updateTooltip()
  tipAnchor = self:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
  tip:SmartAnchorTo(self)
end

function dataobj.OnLeave(self)
  hideTooltip()
end


----------------
--- onClick! ---
----------------
function dataobj.OnClick(self,button)
  if button == "RightButton" then
    return InterfaceOptionsFrame_OpenToCategory(LibStub("AceConfigDialog-3.0").BlizOptions["tomQuest2"].frame)
  else
    ToggleFrame(QuestLogFrame)
  end
end


---------------------
--- Quest Tracker ---
---------------------
function tomQuest2:updateQuestTracker()
  if numTrackedQuests == 0 then return end
  tracker = LibQTip:Acquire("tomQuest2Tracker", 2, "LEFT", "RIGHT")
  tracker:EnableMouse(true)
  tracker:SetBackdropColor(prdb.tracker.backDropColor[1], prdb.tracker.backDropColor[2], prdb.tracker.backDropColor[3], prdb.tracker.backDropColor[4])
  tracker:SetBackdropBorderColor(prdb.tracker.borderColor[1], prdb.tracker.borderColor[2], prdb.tracker.borderColor[3], prdb.tracker.backDropColor[4])
  tracker:SetScale(prdb.tracker.scale)
  tracker:SetFrameStrata("BACKGROUND")

  local headerFont = CreateFont("headerFont")
  headerFont:SetFont(GameTooltipText:GetFont(), prdb.tracker.headerFontSize)
  tracker:SetHeaderFont(headerFont)

  local smallFont = CreateFont("smallFont")
  smallFont:SetFont(GameTooltipText:GetFont(), prdb.tracker.normalFontSize)
  tracker:SetFont(smallFont)

  local tempLine
  for uid in pairs(questsTracked) do
    if uid and quests[uid] then
      local qc = GetDifficultyColor(quests[uid].level)
      local questColor = format("%02x%02x%02x", qc.r*255, qc.g*255, qc.b*255)
      tempLine = tracker:AddLine()
      tracker:SetCell(tempLine, 1, format("|cff%s%s %s|r", questColor, quests[uid].taggedname, quests[uid].playeronquest > 1 and format("<%d>",quests[uid].playeronquest) or ""),headerFont, "LEFT", 1, cProvider, "quest", uid, prdb.tracker.maxwidth, 5, trackerAnchor)
      tracker:SetCell(tempLine, 2, quests[uid].pqstatus, headerFont, "RIGHT", 1, cProvider, "quest", uid, prdb.tracker.maxwidth, 0, trackerAnchor)
      if questsObjectives[uid] then
        for o in pairs(questsObjectives[uid]) do
          tempLine = tracker:AddLine()
          tracker:SetCell(tempLine, 1, format("|cff%s%s|r", questsObjectives[uid][o].qocolor, questsObjectives[uid][o].desc), smallFont, "LEFT", 1, cProvider, "quest", uid, prdb.tooltip.maxwidth, 10, trackerAnchor)
          tracker:SetCell(tempLine, 2, questsObjectives[uid][o].qostatus, smallFont, "RIGHT", 1, cProvider, "quest", uid, prdb.tooltip.maxwidth, 0, trackerAnchor)
        end
      else
        tempLine = tracker:AddLine()
        tracker:SetCell(tempLine, 1, quests[uid].objective, smallFont, "LEFT", 1, cProvider,"quest", uid, prdb.tooltip.maxwidth, 10, trackerAnchor)
      end
    end
  end
  tracker:SmartAnchorTo(anchor)
  tracker:Show()
end